@@ -2,8 +2,9 @@
 import sys
 import yaml
 from . import Container
-from .errors import Error
+from .errors import Error, TestError
 from clank import ArgumentCommand, HelpCommand, Manager, UsageCommand
+from copy import deepcopy
 
 
 def get_containers():
@@ -36,13 +37,18 @@ class InfoCommand(ArgumentCommand):
             data = container.to_dict(self.options.raw)
             if self.options.key:
                 data = data.get(self.options.key)
+            elif data.get('testing'):
+                data['testing'] = deepcopy(data['testing'])
             if isinstance(data, (list, tuple, dict)):
-                sys.stdout.write(yaml.safe_dump(
-                    data, explicit_start=False, default_flow_style=False))
-            else:
+                yaml.safe_dump(data, stream=sys.stdout,
+                               explicit_start=False,
+                               default_flow_style=False)
+            elif isinstance(data, bool):
+                print('true' if data else 'false')
+            elif data is not None:
                 print(data)
         except Error as e:
-            print("error: {}".format(e))
+            print("info failed: {}".format(e))
             return 1
 
 
@@ -73,7 +79,7 @@ class BuildCommand(ArgumentCommand):
                 print("{} pushed".format(container.tag))
             return 0
         except Error as e:
-            print("error: {}".format(e))
+            print("build failed: {}".format(e))
             return 1
 
 
@@ -91,13 +97,13 @@ class PushCommand(ArgumentCommand):
         self.parse_args(args)
         try:
             container = Container.from_path(path=self.options.path, version=self.options.version)
-            if not container.built:
+            if not container.is_built():
                 container.build()
             container.push()
             print("{} pushed".format(container.tag))
             return 0
         except Error as e:
-            print("error: {}".format(e))
+            print("push failed: {}".format(e))
             return 1
 
 
@@ -115,7 +121,7 @@ class RemoveCommand(ArgumentCommand):
         self.parse_args(args)
         try:
             container = Container.from_path(path=self.options.path)
-            if container.built:
+            if container.is_built():
                 container.rm()
                 print ("{} removed".format(container.tag))
                 return 0
@@ -123,7 +129,7 @@ class RemoveCommand(ArgumentCommand):
                 print ("{} not found".format(container.tag))
                 return 2
         except Error as e:
-            print("error: {}".format(e))
+            print("remove failed: {}".format(e))
             return 1
 
 
@@ -143,14 +149,14 @@ class RunCommand(ArgumentCommand):
         self.parse_args(args)
         try:
             container = Container.from_path(path=self.options.path, version=self.options.version)
-            if not container.built:
+            if not container.is_built():
                 container.build()
                 print("{} built".format(container.tag))
             container.run(self.options.detach)
             if self.options.detach:
                 print("{} running".format(container.tag))
         except Error as e:
-            print("error: {}".format(e))
+            print("run failed: {}".format(e))
             return 1
 
 
@@ -169,7 +175,37 @@ class KillCommand(ArgumentCommand):
             container.kill()
             print("{} killed".format(container.tag))
         except Error as e:
-            print("error: {}".format(e))
+            print("kill failed: {}".format(e))
+            return 1
+
+
+class TestCommand(ArgumentCommand):
+    """Test a container."""
+    name = 'test'
+
+    def add_arguments(self):
+        self.argparser.add_argument('-r', '--rebuild', dest='rebuild', action='store_true',
+                                    help="Rebuild the test container before running.")
+        self.argparser.add_argument('-v', '--version', dest='version', nargs=1,
+                                    help="The version of the container to run.")
+        self.argparser.add_argument('path', metavar='PATH', nargs='?',
+                                    help="The path to the container to run.")
+
+    def run(self, args):
+        self.parse_args(args)
+        try:
+            container = Container.from_path(path=self.options.path, version=self.options.version)
+            if not container.is_built():
+                container.build()
+                print("{} built".format(container.tag))
+            try:
+                container.test()
+                print("{} test succeeded".format(container.tag))
+            except TestError as e:
+                print("{} test failed: exit code {}".format(container.tag, e.exit_code))
+                return e.exit_code
+        except Error as e:
+            print("test failed: {}".format(e))
             return 1
 
 
@@ -184,6 +220,7 @@ def run():
             PushCommand,
             RemoveCommand,
             RunCommand,
+            TestCommand,
             UsageCommand,
         ]).run()
     except KeyboardInterrupt:
