@@ -3,14 +3,15 @@ import os
 import pystache
 import subprocess
 import yaml
-from .errors import ConfigError, ContainerError
+from .errors import ConfigError, ContainerError, TestError
 
 config_defaults = {
     'user': None,
     'name': None,
     'prebuild': None,
     'path': None,
-    'fig': None,
+    'running': None,
+    'testing': None,
 }
 
 
@@ -42,8 +43,10 @@ def read_config(configfile):
         if config[key] is not None and not isinstance(config[key], (str, unicode)):
             raise ConfigError("{} must be a string".format(config, key))
 
-    if config['fig'] is not None and not isinstance(config['fig'], dict):
-        raise ConfigError("fig must be a dictionary")
+    if config['running'] is not None and not isinstance(config['running'], dict):
+        raise ConfigError("running must be a dictionary")
+    if config['testing'] is not None and not isinstance(config['testing'], dict):
+        raise ConfigError("running must be a dictionary")
 
     return config
 
@@ -69,23 +72,31 @@ class Container(object):
             the stripped portion exclufing the `/`. This will override any value explicitly set in
             `user`.
         * `prebuild` - A script to run before calling `docker build` on the container.
-        * `fig` - A fig cofiguration to use when running the container.
+        * `running` - A fig cofiguration to use when running the container.
+        * `testing` - Additional fig configuration to use when testing the container. A `test`
+            container must exist for the test command to run.
 
-        The `prebuild` and `fig` values may have values substituted using the Mustache template
-        syntax. Valid substitutions are `name`, `user`, `version`, and `tag`.
+        The `prebuild`, `running`, and `testing` values may have values substituted using the
+        Mustache template syntax. Valid substitutions are `name`, `user`, `version`, and `tag`.
 
         An example may look like this:
 
             name: nginx
             prebuild: ./prebuild {{version}}
-            fig:
+            running:
                 site:
                     image: example/site
                 nginx:
                     image: {{tag}}
                     ports:
                     - 80
-                    volumes_from: site
+                    volumes_from:
+                    - site
+            testing:
+                test:
+                    build: tests
+                    links:
+                    - nginx:www.example.net
 
         Config values may be passed as keyword arguments. In this case they override the values
         provided in the config file.
@@ -120,10 +131,11 @@ class Container(object):
             data.get('name'),
             data.get('version'),
             data.get('prebuild'),
-            data.get('fig'),
+            data.get('running'),
+            data.get('testing'),
         )
 
-    def __init__(self, path, user, name, version=None, prebuild=None, fig=None):
+    def __init__(self, path, user, name, version=None, prebuild=None, running=None, testing=None):
         """
         Initialize the container. The `path` should be a directory containing the container's
         Dockerfile. The `user` and `name` is combined to create the image repo. The `name` should
@@ -136,9 +148,13 @@ class Container(object):
         `prebuild` is `None` and an executable file called `prebuild` exists in `path` it will be
         executed as the prebuild script.
 
-        If the `fig` value is provided it will be written as YAML to be used as the fig
+        If the `running` param is provided it will be written as YAML to be used as the fig
         configuration when running this container. Otherwise a very simple fig.yml file will be
         written to execute the container.
+
+        If the `testing` param is provided it be used to update the `running` value when the test
+        command is called. The test command requires that a container named `test` exists in the
+        `testing` section.
         """
         def aslist(val):
             if isinstance(val, (list, tuple)):
@@ -156,10 +172,13 @@ class Container(object):
 
         self._version = version
         self._prebuild = prebuild
-        self._fig = fig
+        self._running = running
+        self._testing = testing
 
-        if not self.name:
-            raise ContainerError("container has no name")
+        # some CI systems do not allow image removal
+        self.allow_image_remove = True
+        if os.getenv('CIRCLECI', '') == 'true':
+            self.allow_image_remove = False
 
     def to_dict(self, raw=False):
         """
@@ -180,8 +199,10 @@ class Container(object):
                 data['version'] = self._version
             if self._prebuild:
                 data['prebuild'] = self._prebuild
-            if self._fig:
-                data['fig'] = self._fig
+            if self._running:
+                data['running'] = self._running
+            if self._testing:
+                data['testing'] = self._testing
 
             path = rel(cwd, self.path)
             if path and path != '.':
@@ -192,9 +213,10 @@ class Container(object):
                 'user': self.user,
                 'name': self.name,
                 'version': self.version,
-                'fig': self.fig,
-                'built': self.built,
-                'running': self.running,
+                'running': self._get_fig_cfg(False),
+                'testing': self._get_fig_cfg(True),
+                'is-built': self.is_built(),
+                'is-running': self.is_running(),
             }
             path = rel(cwd, self.path)
             if path and path != '.':
@@ -257,8 +279,19 @@ class Container(object):
         from .version import detect
         return detect(self)
 
-    def _get_fig(self):
+    def _get_fig_cfg(self, test):
         """Return the formatted fig configuration."""
+        fig = self._running
+        if not fig:
+            fig = {
+                self.name: {
+                    'image': '{{tag}}',
+                }
+            }
+
+        if test and self._testing:
+            fig.update(self._testing)
+
         def fmt(data):
             res = {}
             for k, v in data.iteritems():
@@ -269,24 +302,27 @@ class Container(object):
                 else:
                     res[k] = v
             return res
-        return fmt(self._fig)
+        return fmt(fig)
 
-    def _write_fig(self):
+    def _write_fig_cfg(self, test):
         """Write the fig configuration to `fig.yml`."""
         configfile = os.path.join(self.path, 'fig.yml')
         try:
             with open(configfile, 'w') as f:
-                fig = self._get_fig()
+                fig = self._get_fig_cfg(test)
                 f.write(yaml.safe_dump(fig, explicit_start=False, default_flow_style=False))
         except IOError:
             raise ContainerError("unable to write to '{}'".format(configfile))
         except yaml.YAMLError as e:
             raise ContainerError("failed to encode config: {}", e)
 
-    @property
-    def fig(self):
-        """Return the formatted fig configuration."""
-        return self._get_fig()
+    def _get_fig_args(self, *args):
+        """
+        Given arguments to fig return the full list of arguments to pass to `subprocess.call`.
+        """
+        project = (self.user or '') + self.name
+        figcfg = os.path.join(self.path, 'fig.yml')
+        return ['fig', '-p', project, '-f', figcfg] + list(args)
 
     @property
     def tag(self):
@@ -304,8 +340,7 @@ class Container(object):
         """Return the expected path to the Dockerfile for this container."""
         return os.path.join(self.path, 'Dockerfile')
 
-    @property
-    def built(self):
+    def is_built(self):
         """
         Return True if the container has been built previously. This does not guarantee that the
         existing image is up to date.
@@ -355,15 +390,14 @@ class Container(object):
             raise ContainerError("{} build failed: no Dockerfile".format(self.tag))
         if prebuild:
             self.prebuild()
+
         args = ['docker', 'build', '-t', self.tag]
-
-        # TODO: remove hack for running under CircleCI.
-        if os.getenv('CIRCLECI', None) is not None:
-            args += ['--force-rm', '--rm']
-
+        if self.allow_image_remove:
+            args += ['--rm', '--force-rm']
         if nocache:
             args.append('--no-cache')
-        args.append('.')
+        args.append(self.path)
+
         try:
             subprocess.check_call(args, cwd=self.path)
         except subprocess.CalledProcessError:
@@ -372,7 +406,7 @@ class Container(object):
 
     def push(self):
         """Push the built container to its registry."""
-        if not self.built:
+        if not self.is_built():
             raise ContainerError("{} push failed: image not built".format(self.name))
         try:
             subprocess.check_call(['docker', 'push', self.tag], cwd=self.path)
@@ -386,13 +420,13 @@ class Container(object):
         try:
             subprocess.check_call(['docker', 'rmi', self.tag], cwd=self.path)
         except subprocess.CalledProcessError:
-            raise ContainerError("{} remove failed")
+            raise ContainerError("{} remove failed".format(self.tag))
 
-    @property
-    def running(self):
+    def is_running(self):
         """Return True if the container is running."""
         try:
-            out = subprocess.check_output(['fig', 'ps', '-q'], cwd=self.path).strip()
+            args = self._get_fig_args('ps', '-q')
+            out = subprocess.check_output(args, cwd=self.path).strip()
             if out:
                 return True
             return False
@@ -406,11 +440,12 @@ class Container(object):
         `fig rm`. If `detach` is set to `True` then the containers will be run in the background.
         They must be stopped by calling `kill` for the `fig rm` cleanup to take place.
         """
-        self._write_fig()
+        self._write_fig_cfg(False)
         try:
-            args = ['fig', 'up']
+            args = ['up']
             if detach:
                 args.append('-d')
+            args = self._get_fig_args(*args)
             subprocess.check_call(args, cwd=self.path)
         except subprocess.CalledProcessError:
             raise ContainerError("{} run failed")
@@ -418,7 +453,8 @@ class Container(object):
             pass
         finally:
             if not detach:
-                subprocess.call(['fig', 'rm', '--force', '-v'], cwd=self.path)
+                args = self._get_fig_args('rm', '--force', '-v')
+                subprocess.call(args, cwd=self.path)
 
     def kill(self):
         """
@@ -426,9 +462,42 @@ class Container(object):
         Once stopped a `fig rm` is issued to clean up the stopp containers.
         """
         try:
-            if self.running:
-                subprocess.check_call(['fig', 'kill'], cwd=self.path)
+            if self.is_running():
+                args = self._get_fig_args('kill')
+                subprocess.check_call(args, cwd=self.path)
         except subprocess.CalledProcessError:
             raise ContainerError("{} kill failed")
         finally:
-            subprocess.call(['fig', 'rm', '--force', '-v'], cwd=self.path)
+            args = self._get_fig_args('rm', '--force', '-v')
+            subprocess.call(args, cwd=self.path)
+
+    def test(self, rebuild=None):
+        """
+        Run the test container. If `rebuild` is `True` the test container will be rebuilt before
+        running.
+
+        Raise `TestError` if the test has an exit code other then 0. Raise ContainerError on all
+        other failures.
+        """
+        if not self._testing or 'test' not in self._testing:
+            raise ContainerError("{} test failed: no test container")
+
+        self._write_fig_cfg(True)
+        if rebuild and 'build' in self._testing['test']:
+            try:
+                args = self._get_fig_args('buid', 'test')
+                subprocess.check_call(args, cwd=self.path)
+            except subprocess.CalledProcessError:
+                msg = "{} test failed: failed to rebuild test container"
+                raise ContainerError(msg.format(self.tag))
+
+        try:
+            args = self._get_fig_args('run', '--rm', 'test')
+            exit_code = subprocess.call(args, cwd=self.path)
+            if exit_code != 0:
+                msg = "{} test failed: returned non-zero exit code {}"
+                raise TestError(msg.format(self.tag, exit_code), exit_code)
+        except KeyboardInterrupt:
+            raise ContainerError("{} test failed: interrupted".format(self.tag))
+        finally:
+            self.kill()
